/*
 * traps.c
 *
 * Copyright (c) 2008 - Roberto Fichera - TeknoSOFT
 *
 * traps.c subsystem for handling backtrace when the application crashes
 * for any question send email to robyf@tekno-soft.it
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>
#include <unistd.h>
#include <signal.h>
#include <execinfo.h>

#include <pthread.h>

#include "traps.h"

typedef struct TrapsSig2Str
{
  int	sig;
  char	*str;
} 
TrapsSig2Str;

static TrapsSig2Str Traps_sig2str[] = {
      	{ 0, 		(char*)"Invalid" },
#if defined (SIGHUP)
        { SIGHUP,	(char*)"SIGHUP"	 },
#endif
#if defined (SIGINT)
        { SIGINT,	(char*)"SIGINT"	 },
#endif
#if defined (SIGQUIT)
        { SIGQUIT,	(char*)"SIGQUIT" },
#endif
#if defined (SIGILL)
        { SIGILL,	(char*)"SIGILL"	 },
#endif
#if defined (SIGTRAP)
        { SIGTRAP,	(char*)"SIGTRAP" },
#endif
#if defined (SIGABRT)
        { SIGABRT,	(char*)"SIGABRT" },
#endif
#if defined (SIGIOT)
        { SIGIOT,	(char*)"SIGIOT"	 },
#endif
#if defined (SIGBUS)
        { SIGBUS,	(char*)"SIGBUS"	 },
#endif
#if defined (SIGFPE)
        { SIGFPE,	(char*)"SIGFPE"	 },
#endif
#if defined (SIGKILL)
        { SIGKILL,	(char*)"SIGKILL" },
#endif
#if defined (SIGUSR1)
        { SIGUSR1,	(char*)"SIGUSR1" },
#endif
#if defined (SIGSEGV)
        { SIGSEGV,	(char*)"SIGSEGV" },
#endif
#if defined (SIGUSR2)
        { SIGUSR2,	(char*)"SIGUSR2" },
#endif
#if defined (SIGPIPE)
        { SIGPIPE,	(char*)"SIGPIPE" },
#endif
#if defined (SIGALRM)
        { SIGALRM,	(char*)"SIGALRM" },
#endif
#if defined (SIGTERM)
        { SIGTERM,	(char*)"SIGTERM" },
#endif
#if defined (SIGSTKFLT)
        { SIGSTKFLT, (char*)"SIGSTKFLT"	 },
#endif
#if defined (SIGCHLD)
        { SIGCHLD,	(char*)"SIGCHLD" },
#endif
#if defined (SIGCONT)
        { SIGCONT,	(char*)"SIGCONT" },
#endif
#if defined (SIGSTOP)
        { SIGSTOP,	(char*)"SIGSTOP" },
#endif
#if defined (SIGTSTP)
        { SIGTSTP,	(char*)"SIGTSTP" },
#endif
#if defined (SIGTTIN)
        { SIGTTIN,	(char*)"SIGTTIN" },
#endif
#if defined (SIGTTOU)
        { SIGTTOU,	(char*)"SIGTTOU" },
#endif
#if defined (SIGURG)
        { SIGURG,	(char*)"SIGURG"	 },
#endif
#if defined (SIGXCPU)
        { SIGXCPU,	(char*)"SIGXCPU" },
#endif
#if defined (SIGXFSZ)
        { SIGXFSZ,	(char*)"SIGXFSZ" },
#endif
#if defined (SIGVTALRM)
        { SIGVTALRM,	(char*)"SIGVTALRM" },
#endif
#if defined (SIGPROF)
        { SIGPROF,	(char*)"SIGPROF" },
#endif
#if defined (SIGWINCH)
        { SIGWINCH,	(char*)"SIGWINCH" },
#endif
#if defined (SIGIO)
        { SIGIO,	(char*)"SIGIO" },
#endif
#if defined (SIGPOLL)
        { SIGPOLL,	(char*)"SIGPOLL" },
#endif
#if defined (SIGPWR)
        { SIGPWR,	(char*)"SIGPWR"	},
#endif
#if defined (SIGSYS)
        { SIGSYS,	(char*)"SIGSYS"	},
#endif
   	{ 0,		(char*)"Not Found"} 
};

/* convert the signal number in string */
static const char* Traps_sigStr( int sig )
{
  int i;

  /* search the signal */
  for (i=0; Traps_sig2str[i].sig != sig; i++)
  {
    /* check for the end of sig2str vector */
    if (i && Traps_sig2str[i].sig == 0)
    {
      break;
    }

    /* nothing */
  }

  /* return the string */
  return( Traps_sig2str[i].str );
}


#define TRAPS_MAX_CALL_TRACE_INFO     64
#define TRAPS_MAX_COREDUMP_SIZE       (2*1024*1024)

/*
 * used to serialize the Traps_sigFault() function
 */
static pthread_mutex_t Traps_sigFaultLock = PTHREAD_MUTEX_INITIALIZER;

/*
 * Obtain a backtrace and print it to stdout.
 */
static void Traps_sigFault( int sig )
{

  void *array[TRAPS_MAX_CALL_TRACE_INFO];
  int size;
  char **strings;
  int i;

  size = backtrace( array, TRAPS_MAX_CALL_TRACE_INFO );
  strings = backtrace_symbols( (void *const *)array, size );

  pthread_mutex_lock( &Traps_sigFaultLock );


  fprintf( stderr, "===============================================================\n" );
  fprintf( stderr, "INTERNAL ERROR: Signal %s (%d) in pid %d\n\n", Traps_sigStr( sig ), sig, (int)getpid () );

  fprintf( stderr, "---------------------------------------------------------------\n" );
  fprintf( stderr, "PID: %d, Obtained %d stack frames.\n\n", (int)getpid(), (int)size );

  for( i = 0; i < size; i++ )
  {
    fprintf( stderr, "%d) %s\n", (int)(size - i), strings[i] );
  }

  fprintf( stderr, "---------------------------------------------------------------\n" );

  fprintf( stderr, "===============================================================\n" );

  pthread_mutex_unlock( &Traps_sigFaultLock );

  free( strings );

  exit( EXIT_FAILURE );
}

/*
 * Block sigs.
 */
static void Traps_blockSignals( int block, int signum )
{
  sigset_t set;

  sigemptyset( &set );
  sigaddset( &set,signum );
  sigprocmask( ( block ? SIG_BLOCK : SIG_UNBLOCK ) ,&set ,NULL );
}

/*
 * Catch a signal. This should implement the following semantics:
 *
 * 1) The handler remains installed after being called.
 * 2) The signal should be blocked during handler execution.
 */
void Traps_catchSignal( int signum, void (*handler)(int ) )
{
  struct sigaction act;

  /* fill with zero */
  memset( &act, 0, sizeof( act ) );

  /* setup the proper handler */
  act.sa_handler = handler;

#ifdef SA_RESTART
  /*
   * We *want* SIGALRM to interrupt a system call. This for compatibility
   * reasons. SunOS doesn't restart by default the interrupted syscall
   */
  if( signum != SIGALRM )
  {
    act.sa_flags = SA_RESTART;
  }
#endif

  /* install the handler */
  sigemptyset( &act.sa_mask );
  sigaddset( &act.sa_mask, signum );
  sigaction( signum, &act, NULL );
}


void Traps_trapSynchronousSignal( void )
{
#ifdef SIGSEGV
  Traps_catchSignal( SIGSEGV, Traps_sigFault );
#endif
#ifdef SIGFPE
  Traps_catchSignal( SIGFPE, Traps_sigFault );
#endif
#ifdef SIGBUS
  Traps_catchSignal( SIGBUS, Traps_sigFault );
#endif
#ifdef SIGILL
  Traps_catchSignal( SIGILL, Traps_sigFault );
#endif
#ifdef SIGQUIT
  Traps_catchSignal( SIGQUIT, Traps_sigFault );
#endif
#ifdef SIGPIPE
  Traps_catchSignal( SIGPIPE, Traps_sigFault );
#endif
#ifdef SIGABRT
  Traps_catchSignal( SIGABRT, Traps_sigFault );
#endif
#ifdef SIGTERM
  Traps_catchSignal( SIGTERM, Traps_sigFault );
#endif
#ifdef SIGKILL
  Traps_catchSignal( SIGKILL, Traps_sigFault );
#endif
}

void Traps_untrapSynchronousSignal( void )
{
#ifdef SIGSEGV
  Traps_catchSignal( SIGSEGV, SIG_DFL );
#endif
#ifdef SIGFPE
  Traps_catchSignal( SIGFPE, SIG_DFL );
#endif
#ifdef SIGBUS
  Traps_catchSignal( SIGBUS, SIG_DFL );
#endif
#ifdef SIGILL
  Traps_catchSignal( SIGILL, SIG_DFL );
#endif
#ifdef SIGQUIT
  Traps_catchSignal( SIGQUIT, SIG_DFL );
#endif
#ifdef SIGPIPE
  Traps_catchSignal( SIGPIPE, SIG_DFL );
#endif
#ifdef SIGABRT
  Traps_catchSignal( SIGABRT, SIG_DFL );
#endif
#ifdef SIGTERM
  Traps_catchSignal( SIGTERM, SIG_DFL );
#endif
#ifdef SIGKILL
  Traps_catchSignal( SIGKILL, SIG_DFL );
#endif
}

